from collections import defaultdict
from enum import Enum, auto
from itertools import product
import random
from functools import reduce
from .DTLOR_DP import check_tip, preorder, postorder, delta_r, find_min_events

## This file includes methods for computing a DTLOR reconciliation graph,
# which represents all optimal reconciliations for a DTLOR instance.
# The graph is implemented as a dictionary, where the keys are Node Tuples, and the values
# are lists of Node Tuples, corresponding to that node's children. Leaf nodes have an empty child list.
# The reconciliation graph consists of three main parts: The Origin part, the DTL part, and the Rearrangement part.
## 1) The Origin part:
# The origin part contains three types of nodes: ORIGIN nodes, LOCATION_MAPPING, and LOCATION_ASSIGNMENT nodes.
# The location mapping nodes in the origin part of the graph always assign a gene node to *, and look like:
# (NodeType.LOCATION_MAPPING, gene_node, "*")
# An origin node indicates that it is optimal for a given node to recieve a real syntenic location via
# an origin event on the branch leading to that node. It looks like:
# (NodeType.ORIGIN, gene_node)
# Finally, a location assignment node is used to choose between multiple optimal possibilities. Each location mapping
# node may have multiple location assignment children, only one of which is used in an optimal reconciliation.
# Each location assignment node holds a mapping for the left and the right child, and looks like
# (NodeType.LOCATION_ASSIGNMENT, left_node, right_node)
# where left_node and right_node are the assignment for the left and the right child respectively.
# In this part of the graph, the left and the right nodes are either origin nodes or location assignment nodes
# that map to syntenic location *.
# Note that even though the node itself is labeled with its left and right children, the preferred way to get
# access to the children of such a node is by indexing the graph to get a list using G[node].
# An origin node has two children, both of which are used in a reconciliation.
# One child is a SPECIES_LIST node whose children are the optimal species mappings for the gene node.
# The second child is a LOCATION_LIST node, whose children are optimal syntenic location mappings for that node.
## 2) The DTL part:
# The DTL part looks like a traditional DTL reconciliation graph. It has six types of nodes: 
# SPECIES_LIST, COSPECIATION, TRANSFER, LOSS, DUPLICATION, SPECIES_MAPPING
# The SPECIES_LIST node, as described above, is the entry point into the DTL graph.
# It loosely corresponds to best_roots from the previous formulation. When a node gets a syntenic location, it can
# be mapped to anywhere on the species tree. The species list keeps track of the optimal species nodes to map a given
# gene node to. Thus, it looks like:
# (NodeType.SPECIES_LIST, gene_node)
# and has SPECIES_MAPPING children, one of which will be chosen in a reconciliation.
# COSPECIATION, TRANSFER, LOSS, and DUPLICATION nodes are the usual DTL event nodes.
# Each has children that are SPECIES_MAPPING nodes. LOSS nodes only have one child. When an event node is part
# of a reconciliation, all of its children are also part of that reconciliation.
# Finally, a SPECIES_MAPPING node records the mapping of a given gene node to a given species node. Because of loss
# events, multiple mapping nodes involving the same gene node on different species nodes may be present in a
# reconciliation. A SPECIES_MAPPING node looks like:
# (NodeType.SPECIES_MAPPING, gene_node, species_node)
## 3) The Rearrangement part:
# The rearrangement part finds the optimal syntenic location mappings for nodes that have a real (non-*) location.
# It has LOCATION_MAPPING, LOCATION_ASSIGNMENT, and LOCATION_LIST nodes.
# A location mapping node records the assignment of a given gene node to a given syntenic location, so it looks like
# (NodeType.LOCATION_MAPPING, gene_node, location)
# This is the same as the way it is used in the Origin part of the graph, except the location no longer *.
# A location assignment node is also used in the same way as the Origin part of the graph:
# it holds the optimal assignment for the left and right children of the mapped node.
# A location list node is introduced in this part of the graph to handle rearrangement events.
# When a node gets a new syntenic location via rearrangement, there may be many optimal locations that it can get
# One of the ways this graph representation saves space is not to represent every possible rearrangement explicitly.
# Instead, a location list node is used to describe every possible optimal (new) syntenic location that each node
# can receive. So a location list looks like:
# (NodeType.LOCATION_LIST, gene_node, [optimal_locations])
# And has many LOCATION_MAPPING children, one of which is chosen when choosing a reconciliation.
## Through origin events, the Origin part of the graph is connected to the DTL and Rearrangement parts of the graph.
# Each Origin node has a SPECIES_LIST child which connects to SPECIES_MAPPING nodes in the DTL part, and it also has
# a LOCATION_LIST child which connects to LOCATION_MAPPING nodes in the rearrangement part.
# For a given origin family, the gene root of that family corresponds to an origin node. To find the rest of the
# reconciliation for that family, one can do the standard reconciliation-finding process starting from
# that origin node. This will yield both a DTL reconciliation that has species mapping nodes, and a synteny
# reconciliation that has location mapping nodes (connected at the top by the origin node).
## Each NodeType has a GraphType, which is used when building a reconciliation.
# For example, DTL event have GraphType.ALL, which means that when building a reconciliation, ALL of their
# children should be included
# On the other hand, SPECIES_MAPPING nodes have GraphType.CHOOSE, because only one of their event node children will
# be used in the reconciliation.
## Finally, there is a graph representation which is slightly less efficient, but represeents R and O events explicitly
# This can be obtained from the original graph using the build_event_graph function.
# In this new representation, O and R events have their own nodes, with only one child. For example, if (g1, l1)
# rearranges to (g2, l2) along the (g1, l1) branch, the corresponding graph will look like:
# { (NodeType.LOCATION_MAPPING, g1, l1) : (NodeType.LOCATION_ASSIGNMENT((.., g2, ...), (.., g3, ...)))
# NodeType.LOCATION_ASSIGNMENT((.., g2, ...), (.., g3, ...), (..., g1, ...)): [(NodeType.REARRANGEMENT, ...)]
# (NodeType.REARRANGEMENT, ...) : [(NodeType.LOCATION_MAPPING, g2, l2)]}
# For the ... in the above example, the rearrangement is labeled with both the parent and the child nodes in order
# to ensure uniqueness.
# This less-efficient graph representation allows computing the median with respect to a metric that involves the
# symmetric set difference between O and R events, rather than the difference between mapping nodes (which can be
# done using the previous representation)
## In order to interact with the graph, the main intended way is to use prune_graph. prune_graph takes an optional
# starting_points parameter, which is the nodes to start from, and it creates a new graph which is only the nodes
# reachable from the given starting points. For example, to get the origin subgraphs for a given graph, we can
# do something like:
# origins = [n for n in graph.keys() if n[0] is NodeType.ORIGIN]
# origin_graphs = [prune_graph(graph, [o]) for o in origins]
# All of the other algorithms (finding medians, etc.) should work on the graphs produced this way.

Infinity = float("inf")

class GraphType(Enum):
    """
    Defines how each NodeType interacts with the graph when
    choosing a reconciliation. Choose is for "mapping" nodes, that
    choose which optimal event below them to use. All is for "event nodes",
    all of whose children must be used if the event is chosen.
    """
    CHOOSE = 1
    ALL = 2

class NodeType(Enum):
    """
    Encodes the type of a node in a reconciliation graph
    """
    # DTL events
    COSPECIATION = (1, GraphType.ALL)
    DUPLICATION = (2, GraphType.ALL)
    LOSS = (3, GraphType.ALL)
    TRANSFER = (4, GraphType.ALL)
    # Assigns locations to the left and right children of a gene node
    LOCATION_ASSIGNMENT = (5, GraphType.ALL)
    # Maps a gene tree node to a list of locations
    LOCATION_LIST = (6, GraphType.CHOOSE)
    # Maps a gene tree node to a syntenic location
    LOCATION_MAPPING = (7, GraphType.CHOOSE)
    # Maps a gene tree node to a list of species tree nodes
    # Used for choosing the optimal root of a reconciliation
    SPECIES_LIST = (8, GraphType.CHOOSE)
    # Maps gene tree node to species tree node
    SPECIES_MAPPING = (9, GraphType.CHOOSE)
    # When a gene gets a true location (which location is yet unknown)
    ORIGIN = (10, GraphType.ALL)
    # "Root" event for choosing the syntenic location of the root
    ROOT = (11, GraphType.CHOOSE)
    # Event node that represents a change from one syntenic location to another
    REARRANGEMENT = (12, GraphType.ALL)
    # Event node that represents the origin of an actual syntenic location
    ORIGIN_EVENT = (13, GraphType.ALL)
    def __init__(self, e_id, graph_type):
        self._e_id = e_id
        self.graph_type = graph_type
    def __repr__(self):
        return "{}".format(self.name)
    def __eq__(self, other):
        return self._e_id == other._e_id
    def __ne__(self, other):
        return self._e_id != other._e_id
    def __hash__(self):
        return hash((self._e_id))

# Weights for each event type for the distance metric
default_event_weights = {t: 1 for t in NodeType}
# Set of nodes which are implicated in the distance metric
# For the location mapping node symmetric set distance
dist_matters_nodes = set([
    NodeType.LOCATION_MAPPING,
    NodeType.DUPLICATION,
    NodeType.LOSS,
    NodeType.TRANSFER,
    NodeType.COSPECIATION])
# For the event symmetric set distance
dist_matters_events = set([
    NodeType.ORIGIN_EVENT,
    NodeType.REARRANGEMENT,
    NodeType.DUPLICATION,
    NodeType.LOSS,
    NodeType.TRANSFER,
    NodeType.COSPECIATION])

def compute_dtlor_graph(species_tree, gene_tree, phi, locus_map, D, T, L, O, R):
    gene_root = next(iter(gene_tree))
    species_root = next(iter(species_tree))
    # First, compute C
    C, C_star, C_graph = DTL_reconcile(species_tree, gene_tree, phi, D, T, L)
    S, S_star, S_graph = synteny_reconcile(species_tree, gene_tree, locus_map, R)
    # Union the two graphs before adding Null events
    G = {**C_graph, **S_graph}
    Origin = {}
    Null = {}
    # Compute the Origin and Null tables
    for eg in postorder(gene_tree):
        _, vp, eg1, eg2 = gene_tree[eg]
        o_cost = C_star[eg][0] + S_star[eg][0] + O
        # Each origin event for gene g
        # leads to a species list node for optimal places to put that gene
        c_choice = (NodeType.SPECIES_LIST, eg)
        # And a location list node for optimal syntenic locations to give it
        s_choice = (NodeType.LOCATION_LIST, eg)
        o_event = (NodeType.ORIGIN, eg)
        G[o_event] = [c_choice, s_choice]
        Origin[eg] = o_cost, [o_event]
        if check_tip(vp, eg1, eg2):
            Null[eg] = Infinity
        else:
            # Null or Origin for left child
            left_null = Null[eg1], [(NodeType.LOCATION_MAPPING, eg1, "*")]
            left_origin = Origin[eg1]
            left_cost, left_nodes = find_min_events([left_null, left_origin])
            # Null or Origin for right child
            right_null = Null[eg2], [(NodeType.LOCATION_MAPPING, eg2, "*")]
            right_origin = Origin[eg2]
            right_cost, right_nodes = find_min_events([right_null, right_origin])
            a_nodes = []
            # Location assignment nodes assign both left and right
            for l_node, r_node in product(left_nodes, right_nodes):
                node = (NodeType.LOCATION_ASSIGNMENT, l_node, r_node)
                G[node] = [l_node, r_node]
                a_nodes.append(node)
            cost = left_cost + right_cost
            Null[eg] = cost
            G[(NodeType.LOCATION_MAPPING, eg, "*")] = a_nodes
    # Compute the final choice node and cost: does the root get a location or "*"?
    root_null = (Null[gene_root], [(NodeType.LOCATION_MAPPING, gene_root, "*")])
    root_origin = Origin[gene_root]
    root_cost, root_events = find_min_events([root_null, root_origin])
    G[(NodeType.ROOT,)] = root_events
    # Remove the non-optimal parts of G
    G = prune_graph(G)
    return root_cost, G

def DTL_reconcile(species_tree, gene_tree, phi, D, T, L):
    A = {}
    # node -> cost
    C = {}
    # node -> [node]
    C_graph = {}
    O = {}
    # Holds the optimal species nodes to map a given gene node
    # gene_node -> (cost, [species_node])
    C_star = {}
    best_switch = {}
    species_root = next(iter(species_tree))
    for eg in postorder(gene_tree):
        C_star[eg] = (Infinity, [])
        _, vp, eg1, eg2 = gene_tree[eg]
        vp_is_a_tip = check_tip(vp, eg1, eg2)
        for es in postorder(species_tree):
            _, vh, es1, es2 = species_tree[es]
            vh_is_a_tip = check_tip(vh, es1, es2)
            # Relevant mapping nodes, for convenience
            eg_es_m = (NodeType.SPECIES_MAPPING, eg, es)
            eg_es1_m = (NodeType.SPECIES_MAPPING, eg, es1)
            eg_es2_m = (NodeType.SPECIES_MAPPING, eg, es2)
            eg1_es_m = (NodeType.SPECIES_MAPPING, eg1, es)
            eg2_es_m = (NodeType.SPECIES_MAPPING, eg2, es)
            # Compute A
            if vh_is_a_tip:
                # Must match the tip mapping phi
                if vp_is_a_tip and phi[eg] == es:
                    A[eg_es_m] = (0, [])
                else:
                    A[eg_es_m] = (Infinity, [])
            else:
                # Cospeciation
                if not vp_is_a_tip:
                    co1_event = (NodeType.COSPECIATION,
                            (NodeType.SPECIES_MAPPING, eg1, es1),
                            (NodeType.SPECIES_MAPPING, eg2, es2))
                    co1_cost = C[(NodeType.SPECIES_MAPPING, eg1, es1)] + \
                            C[(NodeType.SPECIES_MAPPING, eg2, es2)]
                    co1 = (co1_cost, [co1_event])
                    co2_event = (NodeType.COSPECIATION,
                            (NodeType.SPECIES_MAPPING, eg1, es2),
                            (NodeType.SPECIES_MAPPING, eg2, es1))
                    co2_cost = C[(NodeType.SPECIES_MAPPING, eg1, es2)] + \
                            C[(NodeType.SPECIES_MAPPING, eg2, es1)]
                    co2 = (co2_cost, [co2_event])
                    cospeciation = find_min_events([co1, co2])
                else:
                    cospeciation = (Infinity, [])
                # Loss
                loss_es1 = (C[eg_es2_m] + L, [(NodeType.LOSS,
                    eg_es2_m, None)])
                loss_es2 = (C[eg_es1_m] + L, [(NodeType.LOSS,
                    eg_es1_m, None)])
                loss = find_min_events([loss_es1, loss_es2])
                A[eg_es_m] = find_min_events([cospeciation, loss])
            # Compute C
            # Duplication
            if not vp_is_a_tip:
                dup_event = (NodeType.DUPLICATION,
                        eg1_es_m, 
                        eg2_es_m)
                duplication = (D + C[eg1_es_m] + C[eg2_es_m], [dup_event])
            else:
                duplication = (Infinity, [])
            # Transfer
            if not vp_is_a_tip:
                eg2_cost, eg2_locations = best_switch[eg2_es_m]
                eg2_switch_cost = T + C[eg1_es_m] + eg2_cost
                # location[2] extracts the species node from the mapping node that stores the location
                eg2_switch_events = [(NodeType.TRANSFER,
                    (NodeType.SPECIES_MAPPING, eg1, es),
                    (NodeType.SPECIES_MAPPING, eg2, location[2])) \
                        for location in eg2_locations]
                eg2_switch = (eg2_switch_cost, eg2_switch_events)
                eg1_cost, eg1_locations = best_switch[eg1_es_m]
                eg1_switch_cost = T + C[eg2_es_m] + eg1_cost
                eg1_switch_events = [(NodeType.TRANSFER, 
                    (NodeType.SPECIES_MAPPING, eg2, es),
                    (NodeType.SPECIES_MAPPING, eg1, location[2])) \
                        for location in eg1_locations]
                eg1_switch = (eg1_switch_cost, eg1_switch_events)
                transfer = find_min_events([eg2_switch, eg1_switch])
            else:
                transfer = (Infinity, [])
            cost, events = find_min_events([A[eg_es_m], duplication, transfer])
            C[eg_es_m] = cost
            C_graph[eg_es_m] = events
            # Nodes the graph have children which are a list
            # Add the optimal events to the graph with their children
            for event in C_graph[eg_es_m]:
                t,l,r = event
                event_nodes = []
                if l is not None:
                    event_nodes.append(l)
                # Can be none for a loss
                if r is not None:
                    event_nodes.append(r)
                C_graph[event] = event_nodes
            C_star[eg] = find_min_events([C_star[eg], (C[eg_es_m], [es])])
            # Compute O: (cost, [mapping_node])
            O_c = (C[eg_es_m], [eg_es_m])
            if vh_is_a_tip:
                O[eg_es_m] = O_c
            else:
                O[eg_es_m] = find_min_events([O_c, O[eg_es1_m], O[eg_es2_m]])
        # Compute best_switch
        best_switch[(NodeType.SPECIES_MAPPING, eg, species_root)] = (Infinity, [])
        for es in preorder(species_tree):
            _, vh, es1, es2 = species_tree[es]
            vh_is_a_tip = check_tip(vh, es1, es2)
            eg_es_m = (NodeType.SPECIES_MAPPING, eg, es)
            eg_es1_m = (NodeType.SPECIES_MAPPING, eg, es1)
            eg_es2_m = (NodeType.SPECIES_MAPPING, eg, es2)
            if not vh_is_a_tip:
                best_switch[eg_es1_m] = find_min_events([best_switch[eg_es_m], O[eg_es2_m]])
                best_switch[eg_es2_m] = find_min_events([best_switch[eg_es_m], O[eg_es1_m]])
        # Now that we're done computing C_star[eg], add appropriate choice events
        species_choice = (NodeType.SPECIES_LIST, eg)
        C_graph[species_choice] = [(NodeType.SPECIES_MAPPING, eg, es) for es in C_star[eg][1]]
    return C, C_star, C_graph

#TODO gene/species vs. gene/species
def synteny_reconcile(species_tree, gene_tree, locus_map, R):
    # Holds the optimal syntenic locations for a given gene node
    # gene_node -> [location]
    S_star = {}
    # (g,s) -> cost
    S = {}
    # node -> [[node]]
    S_graph = {}
    allsynteny = set(locus_map.values())
    # Capture R for convenience
    delta = lambda l1, l2: delta_r(l1, l2, R)
    for eg in postorder(gene_tree):
        S_star[eg] = (Infinity, [])
        _, vp, eg1, eg2 = gene_tree[eg]
        for lp in allsynteny:
            eg_lp_m = (NodeType.LOCATION_MAPPING, eg, lp)
            if check_tip(vp, eg1, eg2):
                if locus_map[eg] == lp:
                    S[eg_lp_m] = 0
                    S_graph[eg_lp_m] = []
                    # lp is always the optimum for eg with no cost
                    S_star[eg] = (0, [lp])
                else:
                    S[eg_lp_m] = Infinity
            else:
                # Synteny cost for the left child
                l_keep_m = (NodeType.LOCATION_MAPPING, eg1, lp)
                l_keep = (S[l_keep_m], [l_keep_m])
                l_rearrange = (S_star[eg1][0] + R, [(NodeType.LOCATION_LIST, eg1)])
                l_cost, l_nodes = find_min_events([l_keep, l_rearrange])
                # Synteny cost for the right child
                r_keep_m = (NodeType.LOCATION_MAPPING, eg2, lp)
                r_keep = (S[r_keep_m], [r_keep_m])
                r_rearrange = (S_star[eg2][0] + R, [(NodeType.LOCATION_LIST, eg2)])
                r_cost, r_nodes = find_min_events([r_keep, r_rearrange])
                a_nodes = []
                # Create the appropriate assignment nodes
                for l_node, r_node in product(l_nodes, r_nodes):
                    node = (NodeType.LOCATION_ASSIGNMENT, l_node, r_node)
                    S_graph[node] = [l_node, r_node]
                    a_nodes.append(node)
                cost = l_cost + r_cost
                #assert cost < Infinity
                S[(NodeType.LOCATION_MAPPING, eg, lp)] = cost
                S_graph[eg_lp_m] = a_nodes
                S_star[eg] = find_min_events([S_star[eg], (cost, [lp])])
        # Create the appropriate choice node
        location_choice = (NodeType.LOCATION_LIST, eg)
        S_graph[location_choice] = [(NodeType.LOCATION_MAPPING, eg, lp) for lp in S_star[eg][1]]
    return S, S_star, S_graph

#TODO: methods of Graph class?
def prune_graph(G, starting_points=None):
    """
    Removes the unnecessary nodes from G.
    """
    new_G = {}
    # Use a set to prevent reprocessing
    if starting_points is None:
        extant_nodes = set([(NodeType.ROOT,)])
    else:
        for s in starting_points:
            assert s in G
        new_G[(NodeType.ROOT,)] = starting_points
        extant_nodes = set(starting_points)
    while len(extant_nodes) != 0:
        node = extant_nodes.pop()
        if node[0].graph_type is not None:
            children = G[node]
            new_G[node] = children
            extant_nodes |= set(children)
    return new_G

def find_MPR(G, rand=False):
    """
    Samples a traversal from a graph
    """
    MPR = {}
    # BFS
    extant_nodes = [(NodeType.ROOT,)]
    while len(extant_nodes) != 0:
        node = extant_nodes.pop()
        children = G[node]
        if len(children) > 0:
            # Choose or take all children based on the GraphType
            if node[0].graph_type is GraphType.CHOOSE:
                if rand:
                    choice = random.choice(children)
                else:
                    choice = children[0]
                MPR[node] = [choice]
                extant_nodes.append(choice)
            elif node[0].graph_type is GraphType.ALL:
                MPR[node] = children
                extant_nodes.extend(children)
            else:
                assert False, "Bad GraphType"
        else:
            MPR[node] = []
    return MPR

def iter_MPRs(G):
    """
    Iterate over all MPRs in G
    """
    node = (NodeType.ROOT,)
    return iter_MPRs_helper(node, G)

def iter_MPRs_helper(node, G):
    """
    Recursively yield all sub-MPRs in G rooted at node
    """
    children = G[node]
    if len(children) > 0:
        if node[0].graph_type is GraphType.CHOOSE:
            for child in children:
                # Yield a different sub-MPR for each choice of child
                for sub_mpr in iter_MPRs_helper(child, G):
                    mpr = {node: [child]}
                    mpr.update(sub_mpr)
                    yield mpr
        elif node[0].graph_type is GraphType.ALL:
            # Iterator over all possible combinations of MPRs for children
            # Use * operator to unpack list comprehension
            for subcombo in product(*[iter_MPRs_helper(child, G) for child in children]):
                mpr = {node: children}
                # The sub-MPR for each child
                for child_mpr in subcombo:
                    mpr.update(child_mpr)
                yield mpr
        else:
            assert False, "Bad GraphType"
    # Base case: no children
    else:
        yield {node: []}

def graph_search_order(G):
    """
    Iterator over the nodes of G in BFS order (parents before children)
    Uses set, so the order isn't deterministic, but for the bottom-up/top-down
    DP algorithms, this does not matter
    """
    extant_nodes = set([(NodeType.ROOT,)])
    while len(extant_nodes) != 0:
        node = extant_nodes.pop()
        node_children = G[node]
        extant_nodes |= set(node_children)
        yield node

def count_MPRs(G):
    """
    Count the total number of traversals in G (which is the number of optimal MPRs)
    Computes a node -> count table which is the number of optimal sub-MPRs in
    the subtree rooted at node.
    """
    counts = {}
    # Reverse the graph search order to get children before parents
    postorder = list(graph_search_order(G))[::-1]
    for node in postorder:
        children = G[node]
        # Base case: nodes with no children
        if len(children) == 0:
            counts[node] = 1
        else:
            child_counts = [counts[child] for child in children]
            # All means multiply
            if node[0].graph_type is GraphType.ALL:
                counts[node] = reduce(lambda x, y: x * y, child_counts)
            # Choose means add
            elif node[0].graph_type is GraphType.CHOOSE:
                counts[node] = sum(child_counts)
            else:
                assert False, "Bad GraphType"
    return counts

def event_frequencies(G, counts):
    """
    Compute the 'frequency' of every node - the number of MPRs it appears in
    """
    frequencies = defaultdict(int)
    for node in graph_search_order(G):
        if node[0] is NodeType.ROOT:
            frequencies[node] = counts[node]
        # Now compute the frequencies for this node's children
        children = G[node]
        multiplier = frequencies[node] / counts[node]
        for child in children:
            # All passes the frequency along
            if node[0].graph_type is GraphType.ALL:
                frequencies[child] += frequencies[node]
            # Choose divides the frequency according to the relative proportion of the count
            elif node[0].graph_type is GraphType.CHOOSE:
                frequencies[child] += multiplier * counts[child]
            else:
                assert False, "Bad GraphType"
    return frequencies

def median_subgraph(G, frequencies, dist_matters):
    """
    Compute the medians by maximizing the sum of the (pre-adjusted) frequencies
    """
    median_graph = {}
    freq_sums = {}
    # Children to parents
    postorder = list(graph_search_order(G))[::-1]
    for node in postorder:
        children = G[node]
        if len(children) == 0:
            freq_sums[node] = 0
            median_graph[node] = []
        else:
            # Choose the score from the best child
            if node[0].graph_type is GraphType.CHOOSE:
                # Negate them because we want to maximize the freq sum
                child_freq_sums = [(-freq_sums[c], [c]) for c in children]
                # Finding the min events of the negative freq_sums finds the maximum freq_sums
                neg_freq_sum, opt_children = find_min_events(child_freq_sums)
                freq_sums[node] = -neg_freq_sum
                median_graph[node] = opt_children
            # Add the freq_sums of both children and itself
            elif node[0].graph_type is GraphType.ALL:
                median_graph[node] = children
                freq_sums[node] = sum([freq_sums[c] for c in children])
            else:
                assert False, "Bad GraphType"
        if node[0] in dist_matters:
            freq_sums[node] += frequencies[node]
    # Need to prune it because it was built bottom-up, so many suboptimal parts have been added
    median_graph = prune_graph(median_graph)
    return median_graph

#TODO: event weights
def build_median_graph(G, event_weights, dist_matters):
    """
    Build the median graph by doing the 3 necessary DP algorithms
    G: Reconciliation graph
    event_weights: dict from NodeType -> float indicating how much to weight each event
    dist_matters: set of NodeType which indicates which types of nodes are included in the distance metric
    """
    # First, get the counts and node frequencies
    counts = count_MPRs(G)
    freqs = event_frequencies(G, counts)
    # Adjust the frequencies by half to get a median
    adj = 0.5 * counts[(NodeType.ROOT,)]
    adjusted_freqs = {node: event_weights[node[0]] * (freq - adj) for node,freq in freqs.items()}
    return median_subgraph(G, adjusted_freqs, dist_matters)

def build_node_median_graph(G, event_weights=default_event_weights):
    """
    Build the median graph for the location mapping node symmetric set distance
    """
    build_median_graph(G, event_weights, dist_matters_nodes)

def build_event_median_graph(G, event_weights=default_event_weights):
    """
    Build the median graph for the event symmetric set distance
    """
    event_graph = build_event_graph(G)
    return build_median_graph(event_graph, event_weights, dist_matters_events)

def get_mapping_nodes(location_nodes, G):
    """
    Get the list of location mapping nodes below  the given location assignments
    """
    maps = []
    for l in location_nodes:
        if l[0] is NodeType.LOCATION_MAPPING:
            maps.append(l)
        elif l[0] is NodeType.LOCATION_LIST:
            maps.extend(G[l])
        else:
            assert False, "Bad node {}".format(l)
    return maps

def create_r_events(node, G, event_graph):
    """
    Take a location mapping node and create the set of R event nodes and location
    assignments that should appear below it (that replace the location list nodes)
    """
    assignments = []
    children = G[node]
    # Create an assignment node for the left and for the right
    l_nodes = set([location_assignment[1] for location_assignment in children])
    r_nodes = set([location_assignment[2] for location_assignment in children])
    l_maps = get_mapping_nodes(l_nodes, G)
    r_maps = get_mapping_nodes(r_nodes, G)
    # Create the rearrangement events
    for m in l_maps + r_maps:
        if m[2] == node[2]:
            pass
        else:
            r = (NodeType.REARRANGEMENT, node, m)
            event_graph[r] = [m]
    # Create the assignments
    #TODO: doing this product is inefficient, but
    # currently there isn't a NodeType for a mapping node that uses
    # both of its children, and I'm not sure that such a node type
    # should be introduced
    event_graph[node] = []
    for l, r in product(l_maps, r_maps):
        # Add the node to the assignment signature: In this case, two
        # nodes should not share their assignments, since those assignments will
        # have different rearrangement node children
        assignment = (NodeType.LOCATION_ASSIGNMENT, l, r, node)
        if l[2] == node[2]:
            l_node = l
        else:
            l_node = (NodeType.REARRANGEMENT, node, l)
        if r[2] == node[2]:
            r_node = r
        else:
            r_node = (NodeType.REARRANGEMENT, node, r)
        event_graph[assignment] = [l_node, r_node]
        event_graph[node].append(assignment)

def create_o_events(node, G, event_graph):
    """
    Interpolate an Origin event node before each choice of location for an originating gene node
    """
    children = G[node]
    s_choice = children[1]
    root_syntenies = G[s_choice]
    event_graph[s_choice] = []
    for r in root_syntenies:
        origin_event = (NodeType.ORIGIN_EVENT, r)
        event_graph[s_choice].append(origin_event)
        event_graph[origin_event] = [r]
    event_graph[node] = children

def build_event_graph(G):
    """
    Build a graph that explicitly represents R and O events as nodes.
    This graph will be larger than the typical reconciliation graph,
    but is needed for computing medians w.r.t. the event distance.
    """
    event_graph = {}
    for node in graph_search_order(G):
        children = G[node]
        if node[0] is NodeType.LOCATION_MAPPING:
            if node[2] != "*":
                create_r_events(node, G, event_graph)
            # Copy the location assignment nodes
            # This makes it easier to case out copying the location assignments when
            # the syntenic location is not *, since the location assignment node does
            # not keep track of the syntenic location of the parent
            else:
                event_graph[node] = G[node]
                for child in G[node]:
                    event_graph[child] = G[child]
        elif node[0] is NodeType.ORIGIN:
            create_o_events(node, G, event_graph)
        # All location assignment and location list nodes are replaced in the previous cases
        # All other nodes will be faithfully kept
        elif not node[0] is NodeType.LOCATION_ASSIGNMENT and not node[0] is NodeType.LOCATION_LIST:
            event_graph[node] = children
    return event_graph

def get_species_mappings(gene_node, G):
    """
    Helper for get_events that returns all of the species mappings for a given gene
    """
    return [node[2] for node in G if node[0] is NodeType.SPECIES_MAPPING and node[1] == gene_node]

def get_events(MPR):
    """
    Gets a list of events from an MPR. Requires an MPR built from an event graph
    rather than an MPR built from a normal graph. Use build_event_graph(G) to do the conversion first.
    Returns a list of Events. Each DTL event is structured as
    (event_type, gene_node, species_node)
    Where the event_type is a NodeType for DUPLICATION, TRANSFER, LOSS, or COSPECIATION
    Each O and R event is structured as
    (event_type, gene_node, location, [species_node])
    Where the event_type is a NodeType for ORIGIN or REARRANGEMENT
    The list of species nodes are all the locations that the gene_node is mapped to
    (may be singleton, unless there are losses)
    """
    events = []
    for node in graph_search_order(MPR):
        children = MPR[node]
        # Origin events
        if node[0] is NodeType.ORIGIN:
            gene_node = node[1]
            origin_child = MPR[node][1]
            o_event = MPR[origin_child]
            assert len(o_event) == 1, "Invalid MPR: {}".format(o_event)
            origin_child = o_event[0]
            origin_locus = origin_child[1][2]
            species = get_species_mappings(gene_node, MPR)
            event = (NodeType.ORIGIN, gene_node, origin_locus, species)
            events.append(event)
        # DTL events
        elif node[0] is NodeType.SPECIES_MAPPING and len(children) > 0:
            gene_node = node[1]
            species_node = node[2]
            dtl_child = MPR[node]
            assert len(dtl_child) == 1, "Invalid MPR: {}".format(dtl_child)
            dtl_child = dtl_child[0]
            event_type = dtl_child[0]
            event = (event_type, gene_node, species_node)
            events.append(event)
        # Rearrangement events
        elif node[0] is NodeType.LOCATION_MAPPING and node[2] != "*" and len(children) > 0:
            gene_node = node[1]
            assigns = MPR[node]
            assert len(assigns) == 1, "Invalid MPR: {}".format(assigns)
            assign = MPR[assigns[0]]
            assert len(assign) == 2, "Invalid MPR: {}".format(assign)
            l = assign[0]
            r = assign[1]
            for n in [l,r]:
                if n[0] is NodeType.REARRANGEMENT:
                    species = get_species_mappings(n[2][1], MPR)
                    event = (NodeType.REARRANGEMENT, n[2][1], n[2][2], species)
                    events.append(event)
    return events

def score_event(event, d, t, l, o, r):
    if event[0] is NodeType.COSPECIATION:
        return 0
    elif event[0] is NodeType.DUPLICATION:
        return d
    elif event[0] is NodeType.TRANSFER:
        return t
    elif event[0] is NodeType.LOSS:
        return l
    elif event[0] is NodeType.ORIGIN:
        return o
    elif event[0] is NodeType.REARRANGEMENT:
        return r
    else:
        assert False, "Bad Event Type: {}".format(event[0])

def score_events(events_list, d, t, l, o, r):
    return sum(map(lambda x: score_event(x, d, t, l, o, r), events_list))




import sys, os, subprocess, glob
from . import trees
from multiprocessing import Pool
from Bio import Phylo

def runBlast(dbFileL_1,dbFileL_2,paramD):
    '''Run blast comparing every database in dbFileL_1 against every
database in dbFileL_2.
    '''

    # if directory for blast doesn't exist yet, make it
    blastDir = os.path.split(paramD['blastFilePath'])[0]
    if glob.glob(blastDir)==[]:
        os.mkdir(blastDir)

    clineL =  makeBlastClineList(dbFileL_1,dbFileL_2,paramD)

    if clineL != []:
        # we need to run some
        
        # format the databases
        uniqueDbL=list(set(dbFileL_1+dbFileL_2))
        formatDb(uniqueDbL,paramD['blastExecutDirPath'])

        with Pool(processes=paramD['numProcesses']) as p:
            for stderr in p.imap_unordered(subprocessWrapper, clineL):
                pass # ignore stderr

            
def getDbFileL(fastaFilePath,strainNamesT):
    '''Obtains and returns a list of all fasta files that should be run
through blast. Only keeps those that are in stainNamesT.
    '''
    # get list of all fastas in directory
    dbFileL = []
    for dbFile in glob.glob(fastaFilePath):
        dbStem = os.path.split(dbFile)[-1] # strain + _prot.fa or _dna.fa
        if "_prot.fa" in dbStem:
            dbStem = dbStem.split("_prot.fa")[0]
            if dbStem in strainNamesT:
                # this file name is in the tree, include it
                dbFileL.append(dbFile)
    return dbFileL
    
def formatDb(dbFileL,blastExecutDirPath):
    '''Format fasta files in fastaDir for blast by calling the formatdb
executable.'''
    for dbFileName in dbFileL:
        makeblastdbExecutable = os.path.join(blastExecutDirPath,'makeblastdb')
        subprocess.call([makeblastdbExecutable, '-dbtype' ,'prot', '-in', dbFileName],stdout=subprocess.PIPE)
    return

def makeBlastClineList(dbFileL_1, dbFileL_2,paramD):
    '''Create a list of lists, where the sublists have the command line
needed to run blastp on a pair of databases.'''

    blastFilePath = paramD['blastFilePath']
    blastExecutDirPath = paramD['blastExecutDirPath']
    blastFileJoinStr = paramD['blastFileJoinStr']
    
    # catch blast clines from old params files which don't have
    # trailing whitespace (can get rid of this eventually).
    if paramD['blastCLine'][-1] != ' ':
        blastCLine = paramD['blastCLine'] + ' ' + str(paramD['evalueThresh'])
    else:
        blastCLine = paramD['blastCLine'] + str(paramD['evalueThresh'])

    # get a tuple of the blastp command line args for use below
    blastCLineL = processCline(blastCLine)
    blastCLineL[0] = os.path.join(blastExecutDirPath,blastCLineL[0])
    blastCLineT = tuple(blastCLineL)

    # get blast file dir and extension
    splitT = os.path.split(blastFilePath)
    blastDir,rest = splitT
    blastExtension = rest.split("*")[-1]

    # check if fasta files modified more recently than blast files
    shouldBlast = determineShouldBlast(dbFileL_1[0],dbFileL_2[0],blastFilePath,blastFileJoinStr)

    clineL=[]
    for query in dbFileL_1:
        for db in dbFileL_2:

            # split of paths and extensions
            qstem = os.path.split(query)[-1]
            if "_prot.fa" in qstem:
                qstem = qstem.split("_prot.fa")[0]
            else:
                # just remove extension
                qstem = os.path.splitext(qstem)[0]
                
            dbstem = os.path.split(db)[-1]
            if "_prot.fa" in dbstem:
                dbstem = dbstem.split("_prot.fa")[0]
            else:
                dbstem = os.path.splitext(dbstem)[0]
                
            outFN = os.path.join( blastDir, qstem + blastFileJoinStr + dbstem + blastExtension )
                
            # only add to list if this blast file doesn't already exist
            if not os.path.isfile(outFN) or shouldBlast:
                L = list(blastCLineT) + ['-query',query,'-db',db,'-out',outFN]
                clineL.append(L)
                    
    return clineL

def processCline(cline):
    '''Process the command line parameter so that the output format
section can be passed in to subprocess as a unit.
    '''
    # split out the part between quotes
    L = cline.split('"') # " make emacs happy.
    return L[0].split() + [L[1]] + L[2].split()
    
def determineShouldBlast(dbFile1,dbFile2,blastFilePath,blastFileJoinStr):
    '''Compare a blast and a fasta file. If the fasta file was edited
    more recently, return True, otherwise False.
    '''

    dbStem1 = os.path.split(dbFile1)[-1]
    dbStem1 = dbStem1.split("_prot.fa")[0] 

    dbStem2 = os.path.split(dbFile2)[-1]
    dbStem2 = dbStem2.split("_prot.fa")[0] 

    # get dirs and extensions
    blastDir,rest = os.path.split(blastFilePath)
    blastExtension = rest.split("*")[-1]

    fastaFN = dbFile1
    blastFN = os.path.join(blastDir, dbStem1 + blastFileJoinStr + dbStem2 + blastExtension )

    if not os.path.isfile(blastFN):
        # hasn't been run, return True
        return True
    else:
        # times
        fastaModTime = os.path.getmtime(fastaFN)
        blastModTime = os.path.getmtime(blastFN)

        if fastaModTime < blastModTime:
            return False
        else:
            return True

def subprocessWrapper(cline):
    '''A wraper within which we call subprocess. Becasue we're having
blast write to file, we dump std_out. return std_err.'''
    pipes=subprocess.Popen(cline, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = pipes.communicate()
    return stderr

def parseBlastFile(blastFN,evalueThresh,alignCoverThresh,percIdentThresh):
    '''Parse a single blast output file, returning all hits as a list of
tuples. alignCoverThresh is a threshold for the length of the
alignment relative to query and subject length. The files have the
following fields (our custom output) qseqid sseqid evalue qlen qstart
qend slen sstart send.
    '''
    L=[]
    with open(blastFN,'r') as f:
        s = f.readline()
        while s:
            blastLine = s.split()

            # get numeric gene from front, if not present get whole name
            queryStr = blastLine[0].split('_')[0]
            if queryStr.isdigit():
                queryGene = int(queryStr)
            else:
                queryGene = queryStr

            subjectStr = blastLine[1].split('_')[0]
            if subjectStr.isdigit():
                subjectGene = int(subjectStr)
            else:
                subjectGene = subjectStr

            evalue = float(blastLine[2])
            if evalue < evalueThresh:
                qlen = int(blastLine[3])
                qstart = int(blastLine[4])
                qend = int(blastLine[5])
                slen = int(blastLine[6])
                sstart = int(blastLine[7])
                send = int(blastLine[8])
                pident = float(blastLine[9])
                score = int(blastLine[10])

                alCov = ((qend-qstart) + (send-sstart)) / (qlen+slen)
                if alCov > alignCoverThresh and pident > percIdentThresh:
                    L.append((queryGene,subjectGene,evalue,alCov,pident,score))
            s = f.readline()
    return L



# DTLOR_DP.py
# Nuo Liu, HMC 2019-2020 Thesis project
# Modified from Ran Libeskind-Hadas(June 2015)

# The basic DP algorithm for reconciling pairs of trees
# Altered and expanded by Carter Slocum and Annalise Schweickart
# A tree is represented as a dictionary of key-value pairs where a key is an
# edge name and the value is a tuple of the form
# (start vertex, end vertex, left child edge name, right child edge name)
# An edge name may be None.

# Edited by Annalise Schweickart and Carter Slocum, July 2015 to return
# the DTL reconciliation graph that uses frequency scoring, as well as the
# number of reconciliations of the host and parasite trees

# Modified by Ross Mawhorter in 2020 to reduce the running time per Rose Liu's formulation
# and apply the new model of how DTLOR should work.

# import newickFormatReader
from .Greedy import *
import copy
import sys, glob, os
import time
from random import choice

def valid_star(is_top_star, allsynteny):
    """
    returns the set of valid syntenies for descendants
    """
    if is_top_star:
        new_locs = copy.deepcopy(allsynteny)
        new_locs.append("*")
        return new_locs
    else:
        return allsynteny

def valid(synteny, allsynteny):
    """
    returns the set of valid syntenies for the descendents
    """
    is_top_star = (synteny == "*")
    return valid_star(is_top_star, allsynteny)

def delta_cost(synteny1, synteny2, O, R):
    if synteny1==synteny2:
        return 0
    elif synteny1=="*":
        return O
    else:
        return R

# New delta function that does not use O since
# the O costs are calculated separately
def delta_r(synteny1, synteny2, R):
    if synteny1 == synteny2:
        return 0
    else:
        return R

Infinity = float('inf')

def nodes_preorder(tree, root_edge_name):
    """
    Preorder traversal of the /nodes/ of a tree
    """
    top, bottom, left_child, right_child = tree[root_edge_name]
    if left_child is None and right_child is None:
        return [bottom]
    elif left_child is not None and right_child is not None:
        return [bottom] + preorder(tree, left_child) + preorder(tree, right_child)
    else:
        assert False, "Tree with invalid edge: {}".format(root_edge_name)

def nodes_postorder(tree, root_edge_name):
    """
    Postorder traversal of the /nodes/ of a tree
    """
    top, bottom, left_child, right_child = tree[root_edge_name]
    if left_child is None and right_child is None:
        return [bottom]
    elif left_child is not None and right_child is not None:
        return preorder(tree, left_child) + preorder(tree, right_child) + [bottom]
    else:
        assert False, "Tree with invalid edge: {}".format(root_edge_name)

def preorder(tree, rootEdgeName = None):
    """ Takes a tree as input (see format description above) and returns a 
    list of the edges in that tree in preorder (high edges to low edges)"""

    if rootEdgeName is None:
        rootEdgeName = next(iter(tree))

    value = tree[rootEdgeName]
    _,_,leftChildEdgeName,rightChildEdgeName = value

    # base case
    if leftChildEdgeName == None: # then rightChildEdgeName == None also
        return [rootEdgeName]
    else:
        return [rootEdgeName] + \
                preorder(tree, leftChildEdgeName) + \
                preorder(tree, rightChildEdgeName)

def postorder(tree, rootEdgeName = None):
    """ Takes a tree as input (see format description above) and returns a 
    list of the edges in that tree in postorder (low edges to high edges)"""

    if rootEdgeName is None:
        rootEdgeName = next(iter(tree))

    value = tree[rootEdgeName]
    _,_,leftChildEdgeName,rightChildEdgeName = value
    # base case
    if leftChildEdgeName == None: # then rightChildEdgeName == None also
        return [rootEdgeName]
    else:
        return postorder(tree, leftChildEdgeName) + \
               postorder(tree, rightChildEdgeName) + \
               [rootEdgeName]

def check_tip(vh, eh1, eh2):
    if eh1 is None and eh2 is None:
        return True
    elif eh1 is not None and eh2 is not None:
        return False
    else:
        assert False, "Species node with one child: {}".format(vh)

#TODO: rename
def find_min_events(events_list):
    """
    Combines the events list into a single events list
    Each element of events_list is a tuple of (cost, [events])
    Which holds the DP entry for some part of the DP and the associated events
    This combines them into the entry of lowest cost, combining
    all events that have lowest cost
    """
    cost = Infinity
    events = []
    for c,e in events_list:
        if cost > c:
            cost = c
            events = []
        if cost == c:
            events.extend(e)
    return (cost, events)


#TODO: use this instead of list comprehensions
# To do this, replace the list comprehensions in the current code with this
# then use the old find_min_events to combine the output once all the outputs
# are in the (cost, [event]) form
def find_min_events_alt(elements, cost_computer, event_computer):
    """
    Faster alternative to find_min_events.
    Does not compute events for a given cost unless that cost is optimal.
    Also should use way less memory than find_min_events.
    For example, if the elements are [(species_node, location)]
    Then cost_computer should be type:
      (species_node, location) -> float
    and event_computer should be type:
      (species_node, location) -> event
    """
    min_cost = Infinity
    min_events = []
    costs = map(lambda e: cost_computer(*e), elements)
    for element, cost in zip(elements, costs):
        if cost < min_cost:
            min_cost = c
            min_events = []
        if cost == min_cost:
            min_events.extend(event_computer(*element))
    return (min_cost, min_events)

#TODO: optimization where valid returns only the synteny locations for the clade UNDER a given gene node
#TODO: Refactor other code to reflect the tree representation change:
# ("C", None, None) versus ("C", (None, None, None, None), (None, None, None, None))
# and ("L", (...), None) versus ("L", (...), (None, None, None, None)

def DP(host_tree, parasite_tree, phi, locus_map, D, T, L, Origin, R):
    """ Takes a host_tree, parasite_tree, tip mapping function phi, a locus_map, 
        and duplication cost (D), transfer cost (T), loss cost (L), 
        origin cost (O) and rearrange cost(R) and returns the an MPR as a dict.
        The notation and dynamic programming algorithm are explained in the tech report.
        Cospeciation is assumed to cost 0. """

    A = {}  # A, C, O, and best_switch are all defined in tech report
    C = {}
    O = {}
    best_switch = {}
    # All available syntenic locations
    allsynteny = set(locus_map.values())
    # Capture R for ease of use - it never changes
    delta = lambda s1, s2: delta_r(s1, s2, R)
    parasite_root = next(iter(parasite_tree))
    host_root = next(iter(host_tree))
    #print(host_tree)
    #print("The dimensions is %d by %d by %d by %d"%(len(postorder(parasite_tree, parasite_root)),len(Allsynteny), len(Allsynteny),len(postorder(host_tree, host_root))))
    for ep in postorder(parasite_tree, parasite_root):
        _,vp,ep1,ep2 = parasite_tree[ep]
        vp_is_a_tip = check_tip(vp, ep1, ep2)
        for lp in allsynteny:  # The location of ep at the bottom of the branch above ep
            for eh in postorder(host_tree, host_root):
                #print(ep, lp, eh)
                _,vh,eh1,eh2 = host_tree[eh]
                vh_is_a_tip = check_tip(vh, eh1, eh2)
                # Compute A[(ep, eh, lp)]
                if vh_is_a_tip:
                    if vp_is_a_tip and phi[vp] == vh and locus_map[vp]==lp:
                        A[(ep, eh, lp)] = (0, [("C", None, None)])
                    else: 
                        A[(ep, eh, lp)] = (Infinity, [])
                else: # vh is not a tip
                    # Compute cospeciation events
                    if not vp_is_a_tip:
                        def get_cospeciations(l1, l2):
                            synteny_cost = delta(lp, l1) + delta(lp, l2)
                            co1 = (synteny_cost + C[(ep1, eh1, l1)][0] + C[(ep2, eh2, l2)][0], \
                                    [("S", (ep1, eh1, l1), (ep2, eh2, l2))])
                            co2 = (synteny_cost + C[(ep1, eh2, l1)][0] + C[(ep2, eh1, l2)][0], \
                                    [("S", (ep1, eh2, l1), (ep2, eh1, l2))])
                            return find_min_events([co1, co2])
                        cospeciation_list = [get_cospeciations(l1, l2) for l1 in allsynteny for l2 in allsynteny]
                        cospeciations = find_min_events(cospeciation_list)
                    else:
                        cospeciations = (Infinity, [])
                    # Compute loss events
                    # eh1 is the branch where ep is lost
                    loss_eh1 = (C[(ep, eh2, lp)][0] + L, [("L", (vp, eh2, lp), None)])
                    # eh2 is the branch where ep is lost
                    loss_eh2 = (C[(ep, eh1, lp)][0] + L, [("L", (vp, eh1, lp), None)])
                    losses = find_min_events([loss_eh1, loss_eh2])

                    # Determine which event occurs for A[(ep, eh, lp)]
                    A[(ep, eh, lp)] = find_min_events([cospeciations, losses])

                # Compute C[(ep, eh,l_top, lp)]
                # First, compute duplications
                if not vp_is_a_tip:
                    DUPepeh=Infinity
                    # List to keep track of lowest cost duplication event
                    dupList=[Infinity]
                    def get_duplication(l1, l2):
                        synteny_cost = delta(lp, l1) + delta(lp, l2)
                        dup_cost = C[(ep1, eh, l1)][0] + C[(ep2, eh, l2)][0] + D
                        dup_event = ("D", (ep1, vh, l1), (ep2, vh, l2))
                        return (dup_cost, [dup_event])
                    dup_list = [get_duplication(l1, l2) for l1 in allsynteny for l2 in allsynteny]
                    duplications = find_min_events(dup_list)
                else:
                    duplications = (Infinity, [])
               
                # Compute transfer table
                if not vp_is_a_tip:
                    # Over all possible children syntenies
                    def get_transfer(new_l1, new_l2):
                        synteny_cost = delta(lp, new_l1) + delta(lp, new_l2)
                        # Cost to transfer ep2
                        # Transferred child (ep2) keeps the same syntenic location (lp)
                        ep2_cost, ep2_locations = best_switch[(ep2, eh, lp)]
                        ep2_switch_cost = T + synteny_cost + C[(ep1, eh, new_l1)][0] + ep2_cost
                        ep2_switch_events = [("T", (ep1, vh, new_l1), (ep2, location[1], new_l2)) for location in \
                                ep2_locations]
                        ep2_switch = (ep2_switch_cost, ep2_switch_events)
                        # Cost to transfer ep1
                        # Now ep1 is being transferred and keeps lp
                        ep1_cost, ep1_locations = best_switch[(ep1, eh, lp)]
                        ep1_switch_cost = T + synteny_cost + C[(ep2, eh, new_l1)][0] + ep1_cost
                        ep1_switch_events = [("T", (ep2, vh, new_l1), (ep1, location[1], new_l2)) for location in \
                                ep1_locations]
                        ep1_switch = (ep1_switch_cost, ep1_switch_events)
                        return find_min_events([ep2_switch, ep1_switch])
                    transfer_list = [get_transfer(l1, l2) for l1 in allsynteny for l2 in allsynteny]
                    transfers = find_min_events(transfer_list)
                else:
                    transfers = (Infinity, [])

                # Compute C[(ep, eh, l_top, lp)] and the associated events
                C[(ep, eh, lp)] = \
                        find_min_events([A[(ep, eh, lp)], duplications, transfers])
                # The root must factor in the cost of getting a syntenic location
                if ep == parasite_root:
                    old = C[(ep, eh, lp)]
                    C[(ep, eh, lp)] = (old[0] + Origin, old[1])

                # Compute O[(ep, eh, lp)]
                # O is mapping_node -> (cost, [mapping_node])
                if vh_is_a_tip: 
                    O[(ep, eh, lp)] = (C[(ep, eh, lp)][0], [(vp, vh, lp)])
                else: 
                    O_c = (C[(ep, eh, lp)][0], [(vp, vh, lp)])
                    O_eh1 = (O[(ep, eh1, lp)])
                    O_eh2 = (O[(ep, eh2, lp)])
                    O[(ep, eh, lp)] = find_min_events([O_c, O_eh1, O_eh2])

            # Compute best_switch values for the children
            best_switch[(ep, host_root, lp)] = (Infinity, [])
            for eh in preorder(host_tree, host_root):
                _, vh, eh1, eh2 = host_tree[eh]
                # Find the best switches and switch locations for the children of vh
                # Don't set best_switch for nonexistent children
                if not check_tip(vh, eh1, eh2):
                    ep_best_switch = best_switch[(ep, eh, lp)]
                    O_eh2=O[(ep, eh2, lp)]
                    O_eh1=O[(ep, eh1, lp)]
                    best_switch[(ep, eh1, lp)] = find_min_events([ep_best_switch, O_eh2])
                    best_switch[(ep, eh2, lp)] = find_min_events([ep_best_switch, O_eh1])
        # Compute the cost of not giving a syntenic location
        # Tip must have a syntenic location
        if vp_is_a_tip:
            C[(vp, host_root, "*")] = (Infinity, [])
        else:
            def get_single_null(eh, l):
                # Left child stays null
                l_map = (ep1, host_root, "*")
                r_map = (ep2, eh, l)
                left_null_cost = C[l_map][0] + C[r_map][0] + Origin
                left_null_event = ("N", l_map, r_map)
                left_null = (left_null_cost, [left_null_event])
                # Right child stays null
                l_map = (ep1, eh, l)
                r_map = (ep2, host_root, "*")
                right_null_cost = C[l_map][0] + C[r_map][0] + Origin
                right_null_event = ("N", l_map, r_map)
                right_null = (right_null_cost, [right_null_event])
                return find_min_events([left_null, right_null])
            single_null_list = [get_single_null(eh, l) \
                    for eh in postorder(host_tree, host_root) for l in allsynteny]
            single_null = find_min_events(single_null_list)

            # Neither child gets a synteny
            l_map = (ep1, host_root, "*")
            r_map = (ep2, host_root, "*")
            both_null_cost = C[l_map][0] + C[r_map][0]
            both_null_event = ("N", l_map, r_map)
            both_null = (both_null_cost, [both_null_event])

            def get_neither_child_null(eh1, l1, eh2, l2):
                l_map = (ep1, eh1, l1)
                r_map = (ep2, eh2, l2)
                neither_null_cost = C[l_map][0] + C[r_map][0] + 2 * Origin
                neither_null_event = ("N", l_map, r_map)
                return (neither_null_cost, [neither_null_event])

            neither_null_list = [get_neither_child_null(eh1, l1, eh2, l2) \
                    for eh1 in postorder(host_tree, host_root) for eh2 in postorder(host_tree, host_root) \
                    for l1 in allsynteny for l2 in allsynteny]
            neither_null = find_min_events(neither_null_list)

            C[(ep, host_root, "*")] = find_min_events([single_null, both_null, neither_null])

    # Cost for assigning the root a syntenic location
    root_not_null_list = [C[(parasite_root, eh, l)] for eh in postorder(host_tree, host_root) for l in allsynteny]
    # Cost for not assigning a syntenic location
    root_null = C[(parasite_root, host_root, "*")]
    root_list = root_not_null_list + [root_null]
    min_cost, _ = find_min_events(root_list)

    # Find the mapping nodes involving the root of minimum cost
    best_roots = [m for m,c in C.items() if m[0] == parasite_root and c[0] == min_cost]

    # This picks a random MPR from the optimal ones
    MPR = find_MPR(best_roots, C)
    G = MPR_graph(best_roots, C)
    c = count_mprs(best_roots, G)
    return min_cost, MPR, c

def find_MPR(best_roots, C):
    """
    Find a single MPR for the given C dict and best_roots.
    """
    return find_MPR_helper(best_roots, C, {})

def find_MPR_helper(nodes, C, MPR):
    """
    Recursively find a single MPR using C. Does the work for find_MPR.
    """
    mapping = choice(nodes)
    event = choice(C[mapping][1])
    MPR[mapping] = event
    e_type, e_left, e_right = event
    if e_left is not None:
        _ = find_MPR_helper([e_left], C, MPR)
    if e_right is not None:
        _ = find_MPR_helper([e_right], C, MPR)
    return MPR

def MPR_graph(best_roots, C):
    """
    Find the MPR graph for a given C dict and best_roots.
    """
    return MPR_graph_helper(best_roots, C, {})

def MPR_graph_helper(nodes, C, G):
    """
    Recursively create the entire MPR graph. Does the work for MPR_graph.
    """
    for mapping in nodes:
        events = C[mapping][1]
        G[mapping] = events
        for e_type, e_left, e_right in events:
            if e_left is not None:
                MPR_graph_helper([e_left], C, G)
            if e_right is not None:
                MPR_graph_helper([e_right], C, G)
    return G

def count_mprs(best_roots, G):
    """
    Count the number of MPRs for an mpr graph
    """
    return count_mprs_helper(best_roots, G)

def count_mprs_helper(nodes, G):
    totals = []
    for mapping in nodes:
        events = G[mapping]
        e_counts = []
        for e_type, e_left, e_right in events:
            left_count = 1
            if e_left is not None:
                left_count = count_mprs_helper([e_left], G)
            right_count = 1
            if e_right is not None:
                right_count = count_mprs_helper([e_right], G)
            e_count = left_count * right_count
            e_counts.append(e_count)
        total = sum(e_counts)
        totals.append(total)
    return sum(totals)

def preorderDTLORsort(DTLOR, ParasiteRoot):
    """This takes in a DTL reconciliation graph and parasite root and returns 
    a sorted list, orderedKeysL, that is ordered by level from largest to 
    smallest, where level 0 is the root and the highest level has tips."""

    keysL = orderDTLOR(DTLOR, ParasiteRoot)
    uniqueKeysL = sortHelper(DTLOR, keysL)
    orderedKeysL = []
    levelCounter = 0
    while len(orderedKeysL) < len(keysL):
        for mapping in keysL:
            if mapping[-1] == levelCounter:
                orderedKeysL = orderedKeysL + [mapping]
        levelCounter += 1
    
    # lastLevel = orderedKeysL[-1][1]
    return orderedKeysL

def addScores(treeMin, DTLORDict, ScoreDict):
    """Takes the list of reconciliation roots, the DTLOR reconciliation graph, 
    a dictionary of parent nodes, and a dictionary of score values, and 
    returns the DTLOR with the normalized frequency scores calculated."""
    newDTLOR = copy.deepcopy(DTLORDict)
    parentsDict = {}
    ParasiteRoot=treeMin[0][0]
    preOrder = preorderDTLORsort(DTLORDict, ParasiteRoot)
    for root in preOrder:
        #TODO: when would this be true?
        if root != (None, None): #format (key, level)
            vertices = root[0] #format (vp, vh, l, l)
            if root[1] == 0:
                parentsDict[vertices] = ScoreDict[vertices]
            for n in range(len(DTLORDict[vertices])-1): #the last item is always the cost
                _,child1,child2,oldScore = DTLORDict[vertices][n]
                newDTLOR[vertices][n][3] = parentsDict[vertices] * \
                (1.0 * oldScore / ScoreDict[vertices])
                if child1 is not None:
                    if child1 in parentsDict:
                        parentsDict[child1] += newDTLOR[vertices][n][3]
                    else: 
                        parentsDict[child1] = newDTLOR[vertices][n][3] 
                if child2 is not None:
                    if child2 in parentsDict:
                        parentsDict[child2] += newDTLOR[vertices][n][3]
                    else: 
                        parentsDict[child2] = newDTLOR[vertices][n][3]
    normalize = newDTLOR[preOrder[-1][0]][0][-1]  #score of the top most event
    for key in newDTLOR:
        for event in newDTLOR[key][:-1]:
            event[-1] = event[-1]/normalize
    
    return newDTLOR, normalize




import os,sys
from Bio import SeqIO

def parseGenbank(paramD,fastaOutFileDir,genbankFileList,fileNameMapD):
    '''Parse all the genbank (gbff) files in genbankFileList.'''
    
    if genbankFileList == []:
        raise ValueError("List of genbank files to parse is empty.")

    dnaBasedGeneTrees = paramD['dnaBasedGeneTrees']
    
    geneInfoFile = open(paramD['geneInfoFN'], 'w')
    geneOrderOutFile = open(paramD['geneOrderFN'], 'w')
    
    problemGenbankFileL = []
    
    # iterate through list of genbank files
    geneNum =  0 # xenoGI internal gene numbering
    for fileName in genbankFileList:

        geneNum,problemGenbankFileL = parseGenbankSingleFile(geneNum,fileName,dnaBasedGeneTrees,fileNameMapD,geneInfoFile,geneOrderOutFile,fastaOutFileDir,problemGenbankFileL)
        
    geneInfoFile.close()
    geneOrderOutFile.close()

    # If there are any files in problemGenbankFileL, throw error
    if problemGenbankFileL != []:

        with open(paramD['problemGenbankFN'],'w') as problemGenbankF:
            for probT in problemGenbankFileL:
                problemGenbankF.write("\t".join(probT)+"\n")
    
        raise ValueError('Some genbank files have problems with their annotations. They are listed in ' + paramD['problemGenbankFN'] + '. Please remove and run again.\n')
    
def parseGenbankSingleFile(geneNum,fileName,dnaBasedGeneTrees,fileNameMapD,geneInfoFile,geneOrderOutFile,fastaOutFileDir,problemGenbankFileL):
    '''Parse a single genbank file.
    '''
    genbankName = os.path.split(fileName)[-1]
    speciesName = fileNameMapD[genbankName]

    # verify that there are protein annotations
    if not verifyProteinAnnotations(fileName):
        problemGenbankFileL.append((fileName,'lacks protein annotations'))
        return geneNum,problemGenbankFileL 

    # if using dna for gene trees, verify that dna annotations 3x
    # longer than protein
    if dnaBasedGeneTrees and not verifyDnaAnnotations(fileName):
        problemGenbankFileL.append((fileName,'length of dna and protein annotations do not correspond properly'))
        return geneNum,problemGenbankFileL 
    
    # start a block for this species in geneInfoFile
    geneInfoFile.write("# "+speciesName+"\n")
    
    inFile = open(fileName, 'rU')
    protFastaOutName = fastaOutFileDir + speciesName + "_prot.fa"
    protFastaOutFile = open(protFastaOutName, 'w')

    if dnaBasedGeneTrees:
        dnaFastaOutName = fastaOutFileDir + speciesName + '_dna.fa'
        dnaFastaOutFile = open(dnaFastaOutName, 'w')
        
    # start next line in geneOrderOutFile
    geneOrderOutFile.write(speciesName)

    # iterate through chromosomes in the genbank file
    for record in SeqIO.parse(inFile, "genbank"):
        chrom = record.id # .id, as opposed to .name includes the version id
        # iterate through the genes on the chromosome
        genesOnChromL = []
        for feature in record.features:
            # choose only the features that are protein coding genes
            if feature.type == "CDS" and 'protein_id' in feature.qualifiers and 'translation' in feature.qualifiers:

                start = int(feature.location.start)
                end = int(feature.location.end)
                # strand
                if feature.location.strand == 1:
                    strand = "+"
                elif feature.location.strand == -1:
                    strand = "-"
                else:
                    strand = "?"

                aaSeq = feature.qualifiers['translation'][0]
                if dnaBasedGeneTrees:
                    dnaSeq = str(feature.extract(record.seq))

                # common name
                commonName=''
                if 'gene' in feature.qualifiers:
                    commonName=feature.qualifiers['gene'][0]

                # locus tag
                locusTag=''
                if 'locus_tag' in feature.qualifiers:
                    locusTag = feature.qualifiers['locus_tag'][0]

                # protein ID
                proteinId=''
                if 'protein_id' in feature.qualifiers:
                    proteinId = feature.qualifiers['protein_id'][0]

                # get description
                descrip=''
                if 'gene' in feature.qualifiers:
                    descrip += feature.qualifiers['gene'][0]+' - '
                if 'product' in feature.qualifiers:
                    descrip += feature.qualifiers['product'][0]

                # get xenoGI gene name
                if locusTag != '':
                    geneName = str(geneNum) + "_" + speciesName + '-' + locusTag
                else:
                    # if locus tag missing, base name on start and end coordinates
                    geneName = str(geneNum) + "_" + speciesName + '-' + str(start)+"_"+str(end)
                    
                # write to fastaOutFile
                protFastaOutFile.write(">" + geneName + "\n" + aaSeq + "\n")
                if dnaBasedGeneTrees:
                    dnaFastaOutFile.write(">" + geneName + "\n" + dnaSeq + "\n")


                # write to gene information file
                geneInfoFile.write("\t".join([str(geneNum),geneName,commonName,locusTag,proteinId,descrip,chrom,str(start),str(end),strand]) + "\n")
                genesOnChromL.append(str(geneNum))

                geneNum += 1 # increment every time we write a line to geneInfoFile


        if genesOnChromL != []:
            # if not empty, write this chromosome to geneOrderFile                    
            geneOrderOutFile.write("\t"+" ".join(genesOnChromL))

    geneOrderOutFile.write("\n") # add newline to geneOrder file, as we're done with this strain.

    inFile.close()
    protFastaOutFile.close()
    if dnaBasedGeneTrees:
        dnaFastaOutFile.close()

    return geneNum,problemGenbankFileL 

def verifyProteinAnnotations(fileName):
    '''Run through genbank file fileName verifying that it contains
protein annotations. Return True if it does, False if not.'''
    with open(fileName, 'rU') as f:
        for record in SeqIO.parse(f, "genbank"):
            # iterate through the genes on the chromosome
            for feature in record.features:
                # choose only the features that are protein coding genes
                if feature.type == "CDS" and 'protein_id' in feature.qualifiers and 'translation' in feature.qualifiers:
                    # at least one annotation
                    return True
    # made it all the way through with no annotations
    return False
    
def verifyDnaAnnotations(fileName):
    '''Run through genbank file fileName verifying that dna gene
annotations are 3x longer than protein annotations. Return True if so
in all cases, False otherwise.
    '''
    with open(fileName, 'rU') as f:
        for record in SeqIO.parse(f, "genbank"):
            # iterate through the genes on the chromosome
            for feature in record.features:
                # choose only the features that are protein coding genes
                if feature.type == "CDS" and 'protein_id' in feature.qualifiers and 'translation' in feature.qualifiers:
                    aaSeq = feature.qualifiers['translation'][0]
                    dnaSeq = str(feature.extract(record.seq))
                    if (len(aaSeq) + 1) * 3 != len(dnaSeq):
                        return False
    # made it all the way through so annotations ok
    return True
   



import sys
from . import trees,new_DTLOR_DP


class LocusFamily:
    def __init__(self, famNum, locusFamNum, lfMrca,locusNum=None,reconRootKey=None):
        '''Initialize a LocusFamily object with a family number, a LocusFamily
number, and the mrca for this locus family (which may differ from the
mrca for its family.'''
        self.famNum = famNum
        self.locusFamNum = locusFamNum # a unique number for each lf
        self.lfMrca = lfMrca # on species tree
        self.locusNum = locusNum # indicates location, several lf may share
        self.geneD = {}
        # root branch of locus family in family recon. (geneTreeLoc,geneTreeNB)
        self.reconRootKey = reconRootKey

    def addGene(self, gene,genesO):
        '''Add a gene (in numerical form) to this LocusFamily.'''
        strain = genesO.numToStrainName(gene)
        if strain in self.geneD:
            self.geneD[strain].append(gene)
        else:
            self.geneD[strain] = [gene]
        
    def addGenes(self,genesL,genesO):
        '''Add a list (or set etc.) of genes to this LocusFamily.'''
        for gene in genesL:
            self.addGene(gene,genesO)

    def iterGenes(self):
        '''Iterate through all genes in this locus family.'''
        for L in self.geneD.values():
            for gene in L:
                yield gene

    def iterGenesByStrain(self,strain):
        '''Iterate through all genes in this locus family that are in a
        particular strain.
        '''
        if strain in self.geneD:
            for gene in self.geneD[strain]:
                yield gene
        return

    def iterStrains(self):
        '''Iterator that yields all the strains in which this locus family
occurs.'''
        for strain in self.geneD.keys():
            yield strain

    def origin(self,familiesO,rootFocalClade):
        '''Determine the event at the origin of this locus family. Returns "C", "X" or
        or "R". If reconRootKey is None, then no reconcilation has
        been done, and we return " ".

        '''
        if self.reconRootKey == None:
            return " "
        else:
            fam = familiesO.getFamily(self.famNum)
            eventTypeL=[]
            for eventValue in fam.dtlorMprD[self.reconRootKey]:
                eventType = eventValue[0]
                if eventType == 'R' or eventType == 'O':
                    eventTypeL.append(eventType)
            assert len(eventTypeL) < 2, "Should be at most one O or one R on branch."

            if eventTypeL[0] == 'R':
                orig = 'R'
            else:
                orig = fam.origin(familiesO.speciesRtreeO,rootFocalClade)
            
            return orig

    def countEvents(self,familiesO,eventType):
        '''Count the losses in reconciliation associated with this locus family.'''
        fam = familiesO.getFamily(self.famNum)
        # will take reconRootKey as node in counting, since it makes
        # sense to treat O and R events as having happened at very top
        # of their branch.
        if self.reconRootKey == None:
            return None
        else:
            return fam.countEventsBelowNode(self.reconRootKey[0],eventType)

    def dtlorScore(self,familiesO,paramD):
        '''Get dtlor score from reconciliation associated with this locus family.'''
        fam = familiesO.getFamily(self.famNum)
        # will take reconRootKey as node here, since it makes
        # sense to treat O and R events as having happened at very top
        # of their branch.
        if self.reconRootKey == None:
            return None
        else:
            # we need to include the cost for the originating even
            # here because dtlorScoreBelowNode does not include that
            dtlorScore = 0
            if self.origin(familiesO,paramD['rootFocalClade']) in "CX":
                dtlorScore += paramD['originCost']
            elif self.origin(familiesO,paramD['rootFocalClade']) == "R":
                dtlorScore += paramD['rearrangeCost']

            # get scores below
            dtlorScore += fam.dtlorScoreBelowNode(paramD,self.reconRootKey[0])
            
            return dtlorScore
        
    def printReconByGeneTree(self,familiesO,genesO,fileF=sys.stdout):
        '''Print a text summary of the reconciliation corresponding to this
locus family.'''
        if self.reconRootKey == None:
            print(None)
        else:
            fam = familiesO.getFamily(self.famNum)
            rootOfLfSubtree = self.reconRootKey[0]
            lfSubtreeO = fam.geneTreeO.subtree(rootOfLfSubtree)
            fam.printReconByGeneTreeHelper(fam.dtlorMprD,lfSubtreeO,genesO,self.reconRootKey[0],0,fileF)
        #self,dtlorMprD,geneRtreeO,genesO,node,level,fileF=sys.stdout    
    def getStr(self,genesO,sep):
        '''Return a string representation of a single LocusFamily. Separator
between elements given by sep. Elements are: locusFamNum lfMrca gene1
gene2...
        '''
        reconRootKeyStr = ("_".join(map(str,self.reconRootKey)) if self.reconRootKey!=None else str(self.reconRootKey))
        outL=[str(self.locusFamNum),self.lfMrca,str(self.locusNum),reconRootKeyStr]
        
        for geneNum in self.iterGenes():
            outL.append(genesO.numToName(geneNum))

        return sep.join(outL)
    
    def fileStr(self,genesO):
        '''Return a string representation of a single LocusFamily. Format is
        comma separated: 
        '''
        return self.getStr(genesO,',')

    def __repr__(self):
        '''String representation of a LocusFamily, for display purposes.'''
        return "<lf "+str(self.locusFamNum)+">"
        
        
class Family:
    def __init__(self,famNum,mrca,geneTreeO=None,dtlorMprD=None,sourceFam=None):
        '''Base class to be inherited by initialFamily and originFamily.'''

        self.famNum = famNum
        self.mrca = mrca
        self.locusFamiliesL = []   # will contain locusFamily objects for this family
        self.geneTreeO = geneTreeO # gene tree object
        self.dtlorMprD = dtlorMprD # mpr dict
        self.sourceFam = sourceFam # specifies the corresponding fource fam (blast or ifam)
        
    def addLocusFamily(self,lfO):
        self.locusFamiliesL.append(lfO)

    def getLocusFamilies(self):
        return self.locusFamiliesL

    def iterGenes(self):
        '''Iterate through all the genes associated with this family in numerical form.'''
        for lfO in self.getLocusFamilies():
            for gene in lfO.iterGenes():
                yield gene

    def getAllGenes(self):
        '''Collect all the genes present in the LocusFamily objects belonging
to this family. Return as a set.
        '''
        allGenesS=set()
        for lfO in self.getLocusFamilies():
            allGenesS.update(set(lfO.iterGenes()))
        return allGenesS

    def geneCount(self):
        return len(list(self.iterGenes()))
        
    def iterStrains(self):
        '''Iterator that yields all the strains in which this family
occurs.
        '''
        for lfO in self.getLocusFamilies():
            for strain in lfO.iterStrains():
                yield strain
                
    def getOutsideConnections(self,scoresO):
        '''Given a score object, return a set of all outside genes with
connections to this family.'''
        allGenesInFamS = set(self.iterGenes())
        otherGenesS=set()
        for geneNum in allGenesInFamS:
            for otherGene in scoresO.getConnectionsGene(geneNum):
                if not otherGene in allGenesInFamS:
                    otherGenesS.add(otherGene)
        return otherGenesS

    def addGeneTree(self,geneTreeO):
        '''Given tree object geneTreeO, store as attribute.'''
        self.geneTreeO = geneTreeO

    def addMprD(self,mprD):
        '''Given a dictionary mprD with an mpr, store as attribute.
        '''
        self.dtlorMprD = mprD 

    def printReconByGeneTreeHelper(self,dtlorMprD,geneRtreeO,genesO,node,level,fileF=sys.stdout):
        '''Actual recursive function to print reconciliation. Single letter
codes for events are as follows:

        D - duplication
        T - transfer (hgt withing the species tree)
        L - loss (deletion so that gene is lost on species tree lineage)
        O - origin (either core or xeno hgt).
        R - rearrangment event
        '''

        def printOneKey(printPrefix,dtlorMprD,nbKey):
            if nbKey in dtlorMprD:
                for eventT in dtlorMprD[nbKey]:
                    gt,gtNB = nbKey
                    event,stLoc,stNB,locus = eventT
                    outStr = printPrefix+event+" ("+str(gt)+" "+gtNB+") --> "+"("+str(stLoc)+" "+stNB+")"
                    if locus != None:
                        outStr+=" synReg:"+str(locus)
                    print(outStr,file=fileF)

        # recurse over tree, printing out for each branch and node
        levelSpace = "   "*level
        if geneRtreeO.isLeaf(node):
            print(levelSpace+"- "+node+" ["+genesO.numToStrainName(int(node))+"]",file=fileF)
            printOneKey(levelSpace+"  ",dtlorMprD,(node,'b'))
            printOneKey(levelSpace+"  ",dtlorMprD,(node,'n'))
            return
        else:
            childL = []
            for child in geneRtreeO.children(node):
                childL.append(child)
            print(levelSpace+"- "+node,file=fileF)
            printOneKey(levelSpace+"  ",dtlorMprD,(node,'b'))
            printOneKey(levelSpace+"  ",dtlorMprD,(node,'n'))
            for child in childL:
                self.printReconByGeneTreeHelper(dtlorMprD,geneRtreeO,genesO,child,level+1,fileF)

    def fileStr(self,genesO):
        '''Return string representation of single family. Format is: famNum
        <tab> geneTreeO <tab> reconciliation <tab> mrca <tab>
        locusFamNum1,locusFamGenes <tab> locusFamNum2,locusFamGenes...
        The LocusFamily object representations are comma separated.
        The geneTreeO is string version of an Rtree object.
        reconciliation is a string of the dict.
        '''
        # family number and mrca
        outL =[str(self.famNum),self.mrca]

        # geneTreeO
        if self.geneTreeO != None:
            outL.append(self.geneTreeO.fileStr())
        else:
            outL.append("None")

        # dtlorCost
        if self.dtlorCost != None:
            outL.append(str(self.dtlorCost))
        else:
            outL.append("None")

        # dtlorGraphD
        if self.dtlorGraphD != None:
            outL.append(str(self.dtlorGraphD))
        else:
            outL.append("None")

        # dtlorMprD
        if self.dtlorMprD != None:
            outL.append(str(self.dtlorMprD))
        else:
            outL.append("None")

        # sourceFam
        if self.sourceFam != None:
            outL.append(str(self.sourceFam))
        else:
            outL.append("None")

        # productFamT
        if self.productFamT != None:
            outL.append(str(self.productFamT))
        else:
            outL.append("None")
            
        # locus families
        for lfO in self.locusFamiliesL:
            outL.append(lfO.fileStr(genesO))

        return "\t".join(outL)

class initialFamily(Family):

    def __init__(self,famNum,mrca,geneTreeO=None,dtlorCost=None,dtlorGraphD=None,dtlorMprD=None,sourceFam=None,productFamT=None):
        '''Initialize an object of class initialFamily.'''
        super().__init__(famNum,mrca,geneTreeO,dtlorMprD,sourceFam)
        #  dtlorCost is cost from dtlor alg. A positive integer. In
        #  the case where a family was re-run with permissive origin
        #  costs, we make this value negative.
        self.dtlorCost = dtlorCost
        self.dtlorGraphD = dtlorGraphD
        self.productFamT = productFamT # ids of origin families made from this ifam
        
        # for initial families:
        # geneTreeO is presently a rooted tree. Perhaps in future unrooted.
        # dtlorGraphD is the raw DTLOR output, a graph dict.
        # dtlorMprD is the particular MPR subsequently used
        # sourceFam will be the blast family we came from

    def addGraphD(self,graphD):
        '''Given a dictionary graphD with a dtlor graph, store as attribute.
        '''
        self.dtlorGraphD = graphD 
        
    def productOfams(self):
        '''Return a tuple of the origin families that were made from this
initial family. If that hasn't been done yet, returns None.'''
        return self.productFamT
        
    def countMPRs(self):
        '''How many most parsimonious reconcilations are there with the same
optimal cost?

        '''
        return new_DTLOR_DP.count_MPRs(self.dtlorGraphD)[(new_DTLOR_DP.NodeType.ROOT,)]

    def getMprReconDFromGraph(self,speciesPreOrderT,paramD,isMedian,rand):
        '''From the reconciliation graph (dtlor output), get an mpr, convert
to our node based format. Then return the mpr in original dtlor
format, and also our node based format (both dicts). If isMedian is
True, select a median MPR. If rand is True, this will be randomly
chosen. If rand is False, this will be arbitrarily chosen (and the
same each time if done repeatedly).

        '''

        # get event list
        if isMedian:
            eventG = new_DTLOR_DP.build_event_median_graph(self.dtlorGraphD)
        else:
            eventG = new_DTLOR_DP.build_event_graph(self.dtlorGraphD)

        mprOrigFormatD = new_DTLOR_DP.find_MPR(eventG,rand) # one mpr
        mprNodeFormatD = self.__getMprReconDHelper__(mprOrigFormatD,speciesPreOrderT,paramD)
        
        return mprOrigFormatD,mprNodeFormatD

    def iterMprReconDFromGraph(self,speciesPreOrderT,paramD,isMedian):
        '''Same as getMprReconDFromGraph, but iterates over all MPRs.'''

        # get event list
        if isMedian:
            eventG = new_DTLOR_DP.build_event_median_graph(self.dtlorGraphD)
        else:
            eventG = new_DTLOR_DP.build_event_graph(self.dtlorGraphD)

        for mprOrigFormatD in new_DTLOR_DP.iter_MPRs(eventG):
            # iterate over all
            mprNodeFormatD = self.__getMprReconDHelper__(mprOrigFormatD,speciesPreOrderT,paramD)
        
            yield mprOrigFormatD,mprNodeFormatD
        
    def getMprReconDFromMpr(self,speciesPreOrderT,paramD):
        '''Converts the MPR in self.dtlorMprD to node based format and
returns.
        '''
        return self.__getMprReconDHelper__(self.dtlorMprD,speciesPreOrderT,paramD)

    def printAllPossibleReconsFromGraph(self,speciesPreOrderT,paramD,genesO,fileF=sys.stdout):
        '''Iterate over all possible MPRs in the dtlorGraphD and print.'''

        for mprOrigFormatD,mprNodeFormatD in self.iterMprReconDFromGraph(speciesPreOrderT,paramD,False):
            # do all, not just median mprs
            self.printReconByGeneTreeHelper(mprNodeFormatD,self.geneTreeO,genesO,self.geneTreeO.rootNode,0,fileF)
            print("-------")
            print()
            
    def __getMprReconDHelper__(self,mprOrigFormatD,speciesPreOrderT,paramD):
        '''Takes an MPR in the original format of dtlor and converts to our
node based format.
        '''
        ## support funcs
        def getEventTypeStr(eventTypeO):
            '''Convert from the object representation (from dtlor) of an event to
a string.
            '''
            if eventTypeO == new_DTLOR_DP.NodeType.DUPLICATION:
                return "D"
            elif eventTypeO == new_DTLOR_DP.NodeType.TRANSFER:
                return "T"
            elif eventTypeO == new_DTLOR_DP.NodeType.LOSS:
                return "L"
            elif eventTypeO == new_DTLOR_DP.NodeType.ORIGIN:
                return "O"
            elif eventTypeO == new_DTLOR_DP.NodeType.REARRANGEMENT:
                return "R"
            elif eventTypeO == new_DTLOR_DP.NodeType.COSPECIATION:
                return "S"
            else:
                # should never get here
                raise ValueError("Event is not among the D T L O R options.")

        def getSpeciesNodeClosestToTips(speciesTreeLocL,postOrderT):
            for stLoc in postOrderT:
                if stLoc in speciesTreeLocL:
                    return stLoc
            # should never get here
            raise ValueError("Species node not in list of postorder nodes from species tree.")

        def determineDTLNodeOrBranch(eventTypeO):
            '''Determine whether the gene and species locations are on nodes or
        branches based on the event type.'''
            if eventTypeO == new_DTLOR_DP.NodeType.DUPLICATION:
                # implies geneTreeLoc is a node, speciesTreeLoc is a branch
                return ('n','b')
            elif eventTypeO == new_DTLOR_DP.NodeType.TRANSFER:
                # implies geneTreeLoc is a node, speciesTreeLoc is a branch
                return ('n','b')
            elif eventTypeO == new_DTLOR_DP.NodeType.LOSS:
                # implies geneTreeLoc is a branch, speciesTreeLoc is a node
                return ('b','n')
            elif eventTypeO == new_DTLOR_DP.NodeType.COSPECIATION:
                # implies geneTreeLoc is a node, speciesTreeLoc is a node
                return ('n','n')
            else:
                return # should never get here
        
        ## main code
        speciesPostOrderT = speciesPreOrderT[::-1]
                    
        eventL = new_DTLOR_DP.get_events(mprOrigFormatD)

        # create dict with keys (geneTreeLoc,'n/b')
        mprNodeFormatD = {}
        for eventT in eventL:
            if eventT[0] in (new_DTLOR_DP.NodeType.ORIGIN,new_DTLOR_DP.NodeType.REARRANGEMENT):
                (eventTypeO, geneTreeLoc, location, speciesTreeLocL) = eventT
                eventKey = (geneTreeLoc,'b')
                eventType = getEventTypeStr(eventTypeO)
                speciesTreeLoc = getSpeciesNodeClosestToTips(speciesTreeLocL,speciesPostOrderT)
                eventValue = (eventType,speciesTreeLoc,'b',location)
            elif eventT[0] in (new_DTLOR_DP.NodeType.DUPLICATION,new_DTLOR_DP.NodeType.TRANSFER,new_DTLOR_DP.NodeType.LOSS,new_DTLOR_DP.NodeType.COSPECIATION):
                (eventTypeO, geneTreeLoc, speciesTreeLoc) = eventT
                geneTreeNB,speciesTreeNB = determineDTLNodeOrBranch(eventTypeO)
                eventKey = (geneTreeLoc,geneTreeNB)
                eventType = getEventTypeStr(eventTypeO)
                eventValue = (eventType,speciesTreeLoc,speciesTreeNB,None)
                
            # add it
            if eventKey in mprNodeFormatD:
                mprNodeFormatD[eventKey].append(eventValue)
            else:
                mprNodeFormatD[eventKey] = [eventValue]

        # make sure the costs implied by mprNodeFormatD correspond to what
        # dtlor alg gave
        self.__costCheck__(mprNodeFormatD,paramD)
        
        return mprNodeFormatD

    def __costCheck__(self,mprNodeFormatD,paramD):
        '''Sanity check to ensure the events in mprNodeFormatD sum to dtlorCost.'''

        # get costs
        if self.dtlorCost > 0:
            # use normal costs
            D=int(paramD["duplicationCost"])
            T=int(paramD["transferCost"])
            L=int(paramD["lossCost"])
            O=int(paramD["originCost"])
            R=int(paramD["rearrangeCost"])
            dtlorCost = self.dtlorCost
            
        else:
            # neg dtlorCost means we used permissive costs
            D=int(paramD["DTLRcostPermissiveOrigin"])
            T=int(paramD["DTLRcostPermissiveOrigin"])
            L=int(paramD["DTLRcostPermissiveOrigin"])
            O=int(paramD["originCostPermissiveOrigin"])
            R=int(paramD["DTLRcostPermissiveOrigin"])
            dtlorCost = -self.dtlorCost # remove the negative
            
        sm = self.__costSum__(mprNodeFormatD,D,T,L,O,R)
        assert round(sm,3) == round(dtlorCost,3), "Events in this reconcilation don't sum to minCost."

    def __costSum__(self,mprNodeFormatD,D,T,L,O,R):
        '''Sum the costs implied by the recon in mprNodeFormatD.'''
        sm = 0
        for valL in mprNodeFormatD.values():
            for eventType,_,_,_ in valL:
                if eventType == 'D':
                    sm+=D
                elif eventType == 'T':
                    sm+=T
                elif eventType == 'L':
                    sm+=L
                elif eventType == 'O':
                    sm+=O
                elif eventType == 'R':
                    sm+=R
                # S events have no cost
        return sm
    
    def __repr__(self):
        return "<ifam:"+str(self.famNum)+">"

class originFamily(Family):

    def __init__(self,famNum,mrca,geneTreeO=None,dtlorMprD=None,sourceFam=None):
        '''Initialize an object of class originFamily.'''
        super().__init__(famNum,mrca,geneTreeO,dtlorMprD,sourceFam)

        # for origin families:
        # geneTreeO is a rooted tree.
        # dtlorMprD is a converted version of a single DTLOR MPR
        # sourceFam will hold the source ifam for an originFamily

        self.sourceFam = sourceFam # ifam num that originFam came from

        # set to None
        self.dtlorCost = None
        self.dtlorGraphD = None
        self.productFamT = None
        
        # attributes not saved to file and must be recreated when needed
        self.geneHistoryD = None

    def __repr__(self):
        return "<ofam:"+str(self.famNum)+">"

    def createGeneHistoryD(self):
        '''Create a dict of gene histories. Keyed by tip gene. Value is a
string giving history of gene. We convert O events into either C
(core, happened at or before the rootFocalClade) or X (xeno hgt)
events.
        '''
        # support funcs
        def createGeneHistoryStringL(geneRtreeO,node):
            '''Actual recursive function for making strings of gene
histories. Returns a list of tuples. Each tuple contains
(tipGene,historyStr).

            '''
            # get sequence for this branch and node
            hisStr = ""
            for nbKey in [(node,'b'),(node,'n')]:
                if nbKey in self.dtlorMprD:
                    for event,stLoc,stNB,synLocus in self.dtlorMprD[nbKey]:
                        if event in 'LM':
                            pass
                        else:
                            hisStr+=event
                            
            if geneRtreeO.isLeaf(node):
                return [(node,hisStr)]
            else:
                tempL = []
                for child in geneRtreeO.children(node):
                    partialTempL = createGeneHistoryStringL(geneRtreeO,child)
                    tempL.extend(partialTempL)
                    
                # loop and add current hisStr to events in these
                outL = []
                for geneNum,hisStrRestOfTree in tempL:
                    outL.append((geneNum,hisStr+hisStrRestOfTree))

                return outL
        # end support funcs

        # get history strings
        hisStrL = createGeneHistoryStringL(self.geneTreeO,self.geneTreeO.rootNode)

        # need to only have string back to rootFocalClade in species tree. cut it off before that. do this by keeping the species Tree placements.
        
        # put in dict
        self.geneHistoryD = {}
        for geneNum,hisStr in hisStrL:
            self.geneHistoryD[geneNum] = hisStr

    def getGeneHistoryStr(self,geneNum):
        '''Return a gene history string. Possible characters are:

        D - duplication
        T - transfer (hgt within the species tree)
        O - origin event (entry into species tree)
        R - rearrangment event
        S - cospeciation event

        Note, loss (L) events won't show up in these. And we don't
        print the cotermination (M) events.
        '''

        if self.geneHistoryD != None:
            # geneHistoryD already made
            return self.geneHistoryD[geneNum]    
        elif self.dtlorMprD != None:
            # geneHistoryD not created yet and reconciliation is present
            self.createGeneHistoryD()
            return self.geneHistoryD[geneNum]    
        else:
            # this family has no recon, so we can't caculate a history
            return ""

    def origin(self,speciesRtree,rootFocalClade):
        '''Intepret the O event at the base of this family by determining if
it is a core gene (C) or xeno hgt (X). We assess whether it is core or
not at the base of the rootFocalClade of the species tree. If there is
no reconciliation, return empty string.

        '''
        if self.dtlorMprD == None:
            return ""

        # get the origin event out. There should only be one. If more,
        # this is probably an intial family.
        Ocount = 0
        for valL in self.dtlorMprD.values():
            for eventType,stLoc,stNB,synLocus in valL:
                if eventType == 'O':
                    OstLoc = stLoc
                    Ocount+=1
        if Ocount != 1:
            raise ValueError("There should be exactly one origin event in dtlorMprD.")

        # it is C if originated in rfclade branch or ancestors of it
        rfAncT = speciesRtree.ancestors(rootFocalClade)
        if OstLoc in rfAncT:
            # core gene
            orig = 'C'
        else:
            orig = 'X'

        return orig
        
    def printReconByGeneTree(self,genesO,fileF=sys.stdout):
        '''Print a text summary of the reconciliation.'''
        self.printReconByGeneTreeHelper(self.dtlorMprD,self.geneTreeO,genesO,self.geneTreeO.rootNode,0,fileF)

    def countEventsBelowNode(self,geneTreeNode,eventType):
        '''Count how many events of eventType occur in the reconcilation below
geneTreeNode. eventType is a string, D T L O R.'''

        if self.geneTreeO.isLeaf(geneTreeNode):
            return 0
        else:
            eventCtr = 0
            for child in self.geneTreeO.children(geneTreeNode):
                # count on child branch, then recurse
                if (child,'b') in self.dtlorMprD:
                    for eventValue in self.dtlorMprD[(child,'b')]:
                        if eventValue[0] == eventType:
                            eventCtr += 1
                eventCtr += self.countEventsBelowNode(child,eventType)
            return eventCtr

    def dtlorScoreBelowNode(self,paramD,geneTreeNode):
        '''Compile dtlor score in the reconcilation below geneTreeNode.
        '''
        if self.geneTreeO.isLeaf(geneTreeNode):
            return 0
        else:
            dtlorScore = 0
            for child in self.geneTreeO.children(geneTreeNode):
                # count on child branch, then recurse
                if (child,'b') in self.dtlorMprD:
                    for eventValue in self.dtlorMprD[(child,'b')]:
                        if eventValue[0] == 'D':
                            dtlorScore += paramD['duplicationCost']
                        elif  eventValue[0] == 'T':
                            dtlorScore += paramD['transferCost']
                        elif  eventValue[0] == 'L':
                            dtlorScore += paramD['lossCost']
                        elif  eventValue[0] == 'O':
                            dtlorScore += paramD['originCost']
                        elif  eventValue[0] == 'R':
                            dtlorScore += paramD['rearrangeCost']
                dtlorScore += self.dtlorScoreBelowNode(paramD,child)
            return dtlorScore
        
    def getNewickGeneTreeWithReconLabels(self,genesO):
        '''Print out a Newick string where the nodes are labeled with
events. Format for annotations are [branch events | node events ]. For
tips on the gene tree, the "node event" is the species it ends up
in.
        '''
        
        # construct nodeLabelD
        nodeLabelD = {}
        for node in self.geneTreeO.preorder():
            nodeStr=""
            # branch events
            if (node,'b') in self.dtlorMprD:
                for eventT in self.dtlorMprD[(node,'b')]:
                    nodeStr += eventT[0]

            # node events
            nodeStr+="|"
            if (node,'n') in self.dtlorMprD:
                for eventT in self.dtlorMprD[(node,'n')]:
                    nodeStr += eventT[0]
                    
            if self.geneTreeO.isLeaf(node):
                nodeStr += genesO.numToStrainName(int(node))
                    
            nodeLabelD[node] = nodeStr

        # put in newick string
        return self.geneTreeO.toNewickStr(nodeLabelD=nodeLabelD)

class Families:

    def __init__(self,speciesRtreeO):
        '''Initialize an object of class Families.'''

        self.speciesRtreeO = speciesRtreeO
        self.locusFamiliesD = {}
        self.familiesD = {} 

        ## locusFamiliesD has key locusFamNum and value a LocusFamily
        ## object. familiesD has key famNum and value
        ## a Family object.
        
    def initializeFamily(self,famNum,mrca,famType,geneTreeO=None,dtlorCost=None,dtlorGraphD=None,dtlorMprD=None,sourceFam=None,productFamT=None):
        '''Set up an entry for family famNum.'''

        if famType == "initial":
            self.familiesD[famNum] = initialFamily(famNum,mrca,geneTreeO=geneTreeO,dtlorCost=dtlorCost,dtlorGraphD=dtlorGraphD,dtlorMprD=dtlorMprD,sourceFam=sourceFam,productFamT=productFamT)
        else:
            self.familiesD[famNum] = originFamily(famNum,mrca,geneTreeO=geneTreeO,dtlorMprD=dtlorMprD,sourceFam=sourceFam)

    def addLocusFamily(self, lfO):
        '''Add a LocusFamily. Assumes initializeFamily has already been called
to create the corresponding family.
        '''
        self.locusFamiliesD[lfO.locusFamNum] =  lfO
        self.familiesD[lfO.famNum].addLocusFamily(lfO)

    def getLocusFamily(self,locusFamNum):
        return self.locusFamiliesD[locusFamNum]

    def getFamily(self,famNum):
        return self.familiesD[famNum]

    def addGeneTreeToFamily(self,famNum,geneTreeO):
        self.familiesD[famNum].addGeneTree(geneTreeO)

    def addMprToFamily(self,famNum,dtlorMprD):
        self.familiesD[famNum].addMprD(dtlorMprD)

    def iterLocusFamilies(self):
        '''Iterate over all LocusFamilies in order of locusFamNum.
        '''
        maxLocusFamNum = max(self.locusFamiliesD.keys())
        for i in range(maxLocusFamNum+1):
            if i in self.locusFamiliesD:
                yield self.locusFamiliesD[i]
        
    def iterFamilies(self):
        '''Iterate over all Families in order of famNum.'''
        maxFamNum = max(self.familiesD.keys())
        for i in range(maxFamNum+1):
            if i in self.familiesD:
                yield self.familiesD[i]
                
    def getAllGenes(self):
        '''Collect all the genes present in the LocusFamily objects belonging
to this instance of the Families class. Return as a set.
        '''
        allGenesS=set()
        for lfO in self.iterLocusFamilies():
            allGenesS.update(set(lfO.iterGenes()))
        return allGenesS

    def getNumFamilies(self):
        return len(self.familiesD)
    
    def getNumLocusFamilies(self):
        return len(self.locusFamiliesD)

    def labelHardCore(self,aabrhHardCoreL,typeToLabel):
        '''Run through all families or locusFamiles (specified by typeToLabel)
and create aabrhHardCore attributes in each, indicating which if any
correspond to aabrh hard core families. This attribute can be [], or a
list containing aabrh family numbers.'''
        # funcs
        def matchOneAabrh(aabrhInd,aabrhS,flfItr):
            '''Match single aabrh to objects in flfItr. If match, set atribute of
               that object to True. Return True if match, False if none.
            '''
            for flfO in flfItr():
                S = set(flfO.iterGenes())
                I = S.intersection(aabrhS)
                if len(I) == len(aabrhS):
                    flfO.aabrhHardCore.append(aabrhInd)
                    return True
            return False
        
        # labelHardCore main
        if typeToLabel == 'locusFamily':
            flfItr = self.iterLocusFamilies
        elif typeToLabel == 'Family':
            flfItr = self.iterFamilies

        # create aabrhHardCore attribute, and set to None
        for flfO in flfItr():
            flfO.aabrhHardCore = []

        # now match
        for i,aabrhT in enumerate(aabrhHardCoreL):
            aabrhS = set(aabrhT)
            matchOneAabrh(i,aabrhS,flfItr) # records in object attribute aabrhHardCoreB
            #if not matchOneAabrh(i,aabrhS,flfItr):
            #    print("aabrhS",i,"has no match")
                    
    def __repr__(self):
        return "<Families object--"+str(len(self.familiesD))+" Families, "+str(len(self.locusFamiliesD))+" LocusFamilies>"
